/* transformer_4661.c

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*****************************************************************************/

/* This plugin gives a rough approximation of the LF distortion produced
   by an audio transformer.
   
            --------------------------------->--
   Input-->|                                    |
            -->LPF-->3rd harmonic generator-->- + -->Output

   The LPF is a single pole low-pass filter of the form y[n] = A0x[n] + B1y[n-1] .
   It is based on Simple Low Pass Filter by Richard W.E. Furse. 

   The LPF also functions as an anti-alias filter. */

/*****************************************************************************/

#include <math.h>
#include <stdlib.h>
#include <string.h>

/*****************************************************************************/

#include "ladspa.h"

/*****************************************************************************/

#define TR_CUTOFF  0
#define TR_INPUT   1
#define TR_OUTPUT  2

/*****************************************************************************/

/* Instance data */
typedef struct {

  LADSPA_Data m_Fs;
  LADSPA_Data m_2PiOverFs;

  LADSPA_Data m_lpfLastOutput;
  LADSPA_Data m_TransOutput;
  LADSPA_Data m_FcLast;
  LADSPA_Data m_A0; /* coeff of x[n] */
  LADSPA_Data m_B1; /* coeff of y[n-1] */

  /* Ports:
     ------ */

  LADSPA_Data * m_Fc;
  LADSPA_Data * m_Input;
  LADSPA_Data * m_Output;

} Transformer;

/*****************************************************************************/

/* Construct a new plugin instance.*/ 
LADSPA_Handle 
instantiateTransformer(const LADSPA_Descriptor * Descriptor,
			unsigned long             SampleRate) {

  Transformer * psTrans;

  psTrans = (Transformer *)malloc(sizeof(Transformer));

  if (psTrans) {
    psTrans->m_Fs = (LADSPA_Data)SampleRate;
    psTrans->m_2PiOverFs = (2 * M_PI) / (LADSPA_Data)SampleRate;
    psTrans->m_lpfLastOutput = 0;
    psTrans->m_TransOutput = 0;
    psTrans->m_FcLast = 0;
    psTrans->m_A0 = 0;
    psTrans->m_B1 = 0;
  }

  return psTrans;
}

/*****************************************************************************/

/* Initialise and activate a plugin instance.*/ 
void
activateTransformer(LADSPA_Handle Instance) {
  Transformer * psTransformer;
  psTransformer = (Transformer *)Instance;
  psTransformer->m_lpfLastOutput = 0;
}

/*****************************************************************************/

/* Connect a port to a data location.*/
void 
connectPortToTransformer(LADSPA_Handle Instance,
			  unsigned long Port,
			  LADSPA_Data * DataLocation) {
  
  Transformer * psTrans;

  psTrans = (Transformer *)Instance;

  switch (Port) {
  case TR_CUTOFF:
    psTrans->m_Fc = DataLocation;
    break;
  case TR_INPUT:
    psTrans->m_Input = DataLocation;
    break;
  case TR_OUTPUT:
    psTrans->m_Output = DataLocation;
    break;
  }
}

/*****************************************************************************/

/* Run the LPF and 3rd harmonic algorithms for a block of SampleCount samples. */
void 
runTransformer(LADSPA_Handle Instance,
		       unsigned long SampleCount) {

  LADSPA_Data * Input;
  LADSPA_Data * Output;
  LADSPA_Data A0;
  LADSPA_Data B1;
  LADSPA_Data Comp;
  LADSPA_Data lpfLastOutput;
  LADSPA_Data TransOutput;
  Transformer * psTrans;
  unsigned long lSampleIndex;

  psTrans = (Transformer *)Instance;

  Input = psTrans->m_Input;
  Output = psTrans->m_Output;

  if (*psTrans->m_Fc != 
      psTrans->m_FcLast) {
    psTrans->m_FcLast = *psTrans->m_Fc;
    if (psTrans->m_FcLast <= 0) {
      /* Reject everything. */
      psTrans->m_A0 = psTrans->m_B1 = 0;
    }
    else if (psTrans->m_FcLast > psTrans->m_Fs * 0.0065) {
      /* Limit to Fs * ln(B1)/(-2piFc) */
      psTrans->m_A0 = 0.04;
      psTrans->m_B1 = 0.96;
    }
    else {
      psTrans->m_B1 = 0;
      Comp = 2 - cos(psTrans->m_2PiOverFs
		      * psTrans->m_FcLast);
      psTrans->m_B1 = Comp - (LADSPA_Data)sqrt(Comp * Comp - 1);
      psTrans->m_A0 = 1 - psTrans->m_B1;
    }

  }

  A0 = psTrans->m_A0;
  B1 = psTrans->m_B1;
  lpfLastOutput = psTrans->m_lpfLastOutput;
  TransOutput = psTrans->m_TransOutput;

  for (lSampleIndex = 0;
       lSampleIndex < SampleCount;
       lSampleIndex++) {
	LADSPA_Data x = *(Input);
	/* LPF with dc offset to prevent denormals. Offset needs to be big
	   enough to work after it's cubed, so 1e-10 used. */
	lpfLastOutput = (A0 * (*(Input++) + 1e-10) + B1 * lpfLastOutput);
	LADSPA_Data y = lpfLastOutput;
	/* Cube filter output to convert to third harmonic and add makeup
	   gain */
	TransOutput = x + 0.06 * y - y*y*y/8.0;
	*(Output++) = TransOutput;
  }
  
  psTrans->m_lpfLastOutput = lpfLastOutput;
}

/*****************************************************************************/

/* Throw away a filter instance. */
void 
cleanupTransformer(LADSPA_Handle Instance) {
  free(Instance);
}

/*****************************************************************************/

LADSPA_Descriptor * g_psTransDescriptor = NULL;

/*****************************************************************************/

/* _init() is called automatically when the plugin library is first
   loaded. */
void 
_init() {

  char ** pcPortNames;
  LADSPA_PortDescriptor * piPortDescriptors;
  LADSPA_PortRangeHint * psPortRangeHints;
  
  g_psTransDescriptor
    = (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

  if (g_psTransDescriptor != NULL) {
  
    g_psTransDescriptor->UniqueID
      = 4661;
    g_psTransDescriptor->Label
      = strdup("transformer");
    g_psTransDescriptor->Properties
      = LADSPA_PROPERTY_HARD_RT_CAPABLE;
    g_psTransDescriptor->Name 
      = strdup("Transformer");
    g_psTransDescriptor->Maker
      = strdup("John Rigg");
    g_psTransDescriptor->Copyright
      = strdup("GPL");
    g_psTransDescriptor->PortCount
      = 3;
    piPortDescriptors
      = (LADSPA_PortDescriptor *)calloc(3, sizeof(LADSPA_PortDescriptor));
    g_psTransDescriptor->PortDescriptors
      = (const LADSPA_PortDescriptor *)piPortDescriptors;
    piPortDescriptors[TR_CUTOFF]
      = LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
    piPortDescriptors[TR_INPUT]
      = LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
    piPortDescriptors[TR_OUTPUT]
      = LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
    pcPortNames
      = (char **)calloc(3, sizeof(char *));
    g_psTransDescriptor->PortNames 
      = (const char **)pcPortNames;
    pcPortNames[TR_CUTOFF]
      = strdup("LF Distortion Bandwidth (Hz)");
    pcPortNames[TR_INPUT]
      = strdup("Input");
    pcPortNames[TR_OUTPUT]
      = strdup("Output");
    psPortRangeHints = ((LADSPA_PortRangeHint *)
			calloc(3, sizeof(LADSPA_PortRangeHint)));
    g_psTransDescriptor->PortRangeHints
      = (const LADSPA_PortRangeHint *)psPortRangeHints;
    psPortRangeHints[TR_CUTOFF].HintDescriptor
      = (LADSPA_HINT_BOUNDED_BELOW 
	 | LADSPA_HINT_BOUNDED_ABOVE
	 | LADSPA_HINT_SAMPLE_RATE
	 | LADSPA_HINT_LOGARITHMIC
	 | LADSPA_HINT_DEFAULT_100);
    psPortRangeHints[TR_CUTOFF].LowerBound 
      = 0.0002; /* Fs = 1.0 */
    psPortRangeHints[TR_CUTOFF].UpperBound
      = 0.005; /* Fs = 1.0 */
    psPortRangeHints[TR_INPUT].HintDescriptor
      = 0;
    psPortRangeHints[TR_OUTPUT].HintDescriptor
      = 0;
    g_psTransDescriptor->instantiate 
      = instantiateTransformer;
    g_psTransDescriptor->connect_port 
      = connectPortToTransformer;
    g_psTransDescriptor->activate
      = activateTransformer;
    g_psTransDescriptor->run
      = runTransformer;
    g_psTransDescriptor->run_adding
      = NULL;
    g_psTransDescriptor->set_run_adding_gain
      = NULL;
    g_psTransDescriptor->deactivate
      = NULL;
    g_psTransDescriptor->cleanup
      = cleanupTransformer;
  }
    
}
  
/*****************************************************************************/

void
deleteDescriptor(LADSPA_Descriptor * psDescriptor) {
  unsigned long lIndex;
  if (psDescriptor) {
    free((char *)psDescriptor->Label);
    free((char *)psDescriptor->Name);
    free((char *)psDescriptor->Maker);
    free((char *)psDescriptor->Copyright);
    free((LADSPA_PortDescriptor *)psDescriptor->PortDescriptors);
    for (lIndex = 0; lIndex < psDescriptor->PortCount; lIndex++)
      free((char *)(psDescriptor->PortNames[lIndex]));
    free((char **)psDescriptor->PortNames);
    free((LADSPA_PortRangeHint *)psDescriptor->PortRangeHints);
    free(psDescriptor);
  }
}

/*****************************************************************************/

/* _fini() is called automatically when the library is unloaded. */
void
_fini() {
  deleteDescriptor(g_psTransDescriptor);
}

/*****************************************************************************/

/* Return a descriptor of the requested plugin type. */
const LADSPA_Descriptor * 
ladspa_descriptor(unsigned long Index) {
  /* Return the requested descriptor or null if the index is out of
     range. */
  switch (Index) {
  case 0:
    return g_psTransDescriptor;
  default:
    return NULL;
  }
}

/*****************************************************************************/

/* EOF */
